/*
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2013 KYOCERA Corporation
 */



#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <linux/i2c-cam-algo-bit.h>

#ifdef DEBUG
#define bit_dbg(level, dev, format, args...) \
	do { \
		if (i2c_debug >= level) \
			dev_dbg(dev, format, ##args); \
	} while (0)
#else
#define bit_dbg(level, dev, format, args...) \
	do {} while (0)
#endif

static int bit_test;	
module_param(bit_test, bool, 0);
MODULE_PARM_DESC(bit_test, "Test the lines of the bus to see if it is stuck");

#ifdef DEBUG
static int i2c_debug = 1;
module_param(i2c_debug, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(i2c_debug,
		 "debug level - 0 off; 1 normal; 2 verbose; 3 very verbose");
#endif

#define setsda(adap, val)	adap->setsda(adap->data, val)
#define setscl(adap, val)	adap->setscl(adap->data, val)
#define getsda(adap)		adap->getsda(adap->data)
#define getscl(adap)		adap->getscl(adap->data)

static int sclhi(struct i2c_algo_bit_data *adap)
{
	unsigned long start;

	setscl(adap, 1);

	
	if (!adap->getscl)
		goto done;

	start = jiffies;
	while (!getscl(adap)) {
		
		if (time_after(jiffies, start + adap->timeout))
			return -ETIMEDOUT;
		cond_resched();
	}
#ifdef DEBUG
	if (jiffies != start && i2c_debug >= 3)
		pr_debug("i2c-algo-bit: needed %ld jiffies for SCL to go "
			 "high\n", jiffies - start);
#endif

done:

	return 0;
}

static void i2c_repstart(struct i2c_algo_bit_data *adap)
{
	
	setsda(adap, 1);
	sclhi(adap);
	setsda(adap, 0);

	setscl(adap, 0);
}

static int i2c_outb(struct i2c_adapter *i2c_adap, unsigned char c)
{

	int sb;
	int ack;
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;

	

        sb = (c >> 7) & 1;
		setsda(adap, sb);
		if (sclhi(adap) < 0) {
			bit_dbg(1, &i2c_adap->dev, "i2c_outb: 0x%02x, "
				"timeout at bit #%d\n", (int)c, i);
			return -ETIMEDOUT;
		}
		
		setscl(adap, 0);

        sb = (c >> 6) & 1;
        setsda(adap, sb);
        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_outb: 0x%02x, "
                "timeout at bit #%d\n", (int)c, i);
            return -ETIMEDOUT;
        }

        setscl(adap, 0);

        sb = (c >> 5) & 1;
        setsda(adap, sb);
        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_outb: 0x%02x, "
                "timeout at bit #%d\n", (int)c, i);
            return -ETIMEDOUT;
        }

        setscl(adap, 0);

        sb = (c >> 4) & 1;
        setsda(adap, sb);
        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_outb: 0x%02x, "
                "timeout at bit #%d\n", (int)c, 4);
            return -ETIMEDOUT;
        }

        setscl(adap, 0);

        sb = (c >> 3) & 1;
        setsda(adap, sb);

        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_outb: 0x%02x, "
                "timeout at bit #%d\n", (int)c, 3);
            return -ETIMEDOUT;
        }

        setscl(adap, 0);

        sb = (c >> 2) & 1;
        setsda(adap, sb);
        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_outb: 0x%02x, "
                "timeout at bit #%d\n", (int)c, 2);
            return -ETIMEDOUT;
        }

        setscl(adap, 0);

        sb = (c >> 1) & 1;
        setsda(adap, sb);
        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_outb: 0x%02x, "
                "timeout at bit #%d\n", (int)c, 1);
            return -ETIMEDOUT;
        }

        setscl(adap, 0);

        sb = (c >> 0) & 1;
        setsda(adap, sb);
        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_outb: 0x%02x, "
                "timeout at bit #%d\n", (int)c, 0);
            return -ETIMEDOUT;
        }

        setscl(adap, 0);

	setsda(adap, 1);
	if (sclhi(adap) < 0) {
		bit_dbg(1, &i2c_adap->dev, "i2c_outb: 0x%02x, "
			"timeout at ack\n", (int)c);
		return -ETIMEDOUT;
	}

	

    while(!getscl(adap));

	ack = !getsda(adap);
	bit_dbg(2, &i2c_adap->dev, "i2c_outb: 0x%02x %s\n", (int)c,
		ack ? "A" : "NA");

	setscl(adap, 0);
	return ack;
	
}

static int i2c_inb(struct i2c_adapter *i2c_adap)
{
	
	

	unsigned char indata = 0;
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;

	
	setsda(adap, 1);

		if (sclhi(adap) < 0) {
			bit_dbg(1, &i2c_adap->dev, "i2c_inb: timeout at bit "
                "#%d\n", 7 );
			return -ETIMEDOUT;
		}
		indata *= 2;
		if (getsda(adap))
			indata |= 0x01;
		setscl(adap, 0);

        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_inb: timeout at bit "
                "#%d\n", 6 );
            return -ETIMEDOUT;
	    }
        indata *= 2;
        if (getsda(adap))
            indata |= 0x01;
        setscl(adap, 0);

        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_inb: timeout at bit "
                "#%d\n", 5);
            return -ETIMEDOUT;
        }
        indata *= 2;
        if (getsda(adap))
            indata |= 0x01;
        setscl(adap, 0);

        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_inb: timeout at bit "
                "#%d\n", 4);
            return -ETIMEDOUT;
        }
        indata *= 2;
        if (getsda(adap))
            indata |= 0x01;
        setscl(adap, 0);

        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_inb: timeout at bit "
                "#%d\n", 3);
            return -ETIMEDOUT;
        }
        indata *= 2;
        if (getsda(adap))
            indata |= 0x01;
        setscl(adap, 0);

        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_inb: timeout at bit "
                "#%d\n", 2);
            return -ETIMEDOUT;
        }
        indata *= 2;
        if (getsda(adap))
            indata |= 0x01;
        setscl(adap, 0);

        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_inb: timeout at bit "
                "#%d\n", 1);
            return -ETIMEDOUT;
        }
        indata *= 2;
        if (getsda(adap))
            indata |= 0x01;
        setscl(adap, 0);

        if (sclhi(adap) < 0) {
            bit_dbg(1, &i2c_adap->dev, "i2c_inb: timeout at bit "
                "#%d\n", 0);
            return -ETIMEDOUT;
        }
        indata *= 2;
        if (getsda(adap))
            indata |= 0x01;
        setscl(adap, 0);

	return indata;
}

static int test_bus(struct i2c_algo_bit_data *adap, char *name)
{
	int scl, sda;

	if (adap->getscl == NULL)
		pr_info("%s: Testing SDA only, SCL is not readable\n", name);

	sda = getsda(adap);
	scl = (adap->getscl == NULL) ? 1 : getscl(adap);
	if (!scl || !sda) {
		printk(KERN_WARNING "%s: bus seems to be busy\n", name);
		goto bailout;
	}

	setsda(adap, 0);
	sda = getsda(adap);
	scl = (adap->getscl == NULL) ? 1 : getscl(adap);
	if (sda) {
		printk(KERN_WARNING "%s: SDA stuck high!\n", name);
		goto bailout;
	}
	if (!scl) {
		printk(KERN_WARNING "%s: SCL unexpected low "
		       "while pulling SDA low!\n", name);
		goto bailout;
	}

	setsda(adap, 1);
	sda = getsda(adap);
	scl = (adap->getscl == NULL) ? 1 : getscl(adap);
	if (!sda) {
		printk(KERN_WARNING "%s: SDA stuck low!\n", name);
		goto bailout;
	}
	if (!scl) {
		printk(KERN_WARNING "%s: SCL unexpected low "
		       "while pulling SDA high!\n", name);
		goto bailout;
	}

	setscl(adap, 0);
	sda = getsda(adap);
	scl = (adap->getscl == NULL) ? 0 : getscl(adap);
	if (scl) {
		printk(KERN_WARNING "%s: SCL stuck high!\n", name);
		goto bailout;
	}
	if (!sda) {
		printk(KERN_WARNING "%s: SDA unexpected low "
		       "while pulling SCL low!\n", name);
		goto bailout;
	}

	sclhi(adap);
	sda = getsda(adap);
	scl = (adap->getscl == NULL) ? 1 : getscl(adap);
	if (!scl) {
		printk(KERN_WARNING "%s: SCL stuck low!\n", name);
		goto bailout;
	}
	if (!sda) {
		printk(KERN_WARNING "%s: SDA unexpected low "
		       "while pulling SCL high!\n", name);
		goto bailout;
	}
	pr_info("%s: Test OK\n", name);
	return 0;
bailout:
	setsda(adap, 1);
	sclhi(adap);
	return -ENODEV;
}

static int try_address(struct i2c_adapter *i2c_adap,
		       unsigned char addr, int retries)
{
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;
	int i, ret = 0;

	for (i = 0; i <= retries; i++) {
		ret = i2c_outb(i2c_adap, addr);
		if (ret == 1 || i == retries)
			break;

        setsda(adap, 0);
        sclhi(adap);
        setsda(adap, 1);

		yield();

        setsda(adap, 0);
        setscl(adap, 0);

	}
	if (i && ret)
		bit_dbg(1, &i2c_adap->dev, "Used %d tries to %s client at "
			"0x%02x: %s\n", i + 1,
			addr & 1 ? "read from" : "write to", addr >> 1,
			ret == 1 ? "success" : "failed, timeout?");
	return ret;
}

static int sendbytes(struct i2c_adapter *i2c_adap, struct i2c_msg *msg)
{
	const unsigned char *temp = msg->buf;
	int count = msg->len;
	unsigned short nak_ok = msg->flags & I2C_M_IGNORE_NAK;
	int retval;
	int wrcount = 0;

	while (count > 0) {
		retval = i2c_outb(i2c_adap, *temp);

		
		if ((retval > 0) || (nak_ok && (retval == 0))) {
			count--;
			temp++;
			wrcount++;

		
		} else if (retval == 0) {
			dev_err(&i2c_adap->dev, "sendbytes: NAK bailout.\n");
			return -EIO;

		
		} else {
			dev_err(&i2c_adap->dev, "sendbytes: error %d\n",
					retval);
			return retval;
		}
	}
	return wrcount;
}

static int acknak(struct i2c_adapter *i2c_adap, int is_ack)
{
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;

	
	if (is_ack)		
		setsda(adap, 0);
	if (sclhi(adap) < 0) {	
		dev_err(&i2c_adap->dev, "readbytes: ack/nak timeout\n");
		return -ETIMEDOUT;
	}
	setscl(adap, 0);
	return 0;
}

static int readbytes(struct i2c_adapter *i2c_adap, struct i2c_msg *msg)
{
	int inval;
	int rdcount = 0;	
	unsigned char *temp = msg->buf;
	int count = msg->len;
	const unsigned flags = msg->flags;

	while (count > 0) {
		inval = i2c_inb(i2c_adap);
		if (inval >= 0) {
			*temp = inval;
			rdcount++;
		} else {
			break;
		}

		temp++;
		count--;

		
		if (rdcount == 1 && (flags & I2C_M_RECV_LEN)) {
			if (inval <= 0 || inval > I2C_SMBUS_BLOCK_MAX) {
				if (!(flags & I2C_M_NO_RD_ACK))
					acknak(i2c_adap, 0);
				dev_err(&i2c_adap->dev, "readbytes: invalid "
					"block length (%d)\n", inval);
				return -EREMOTEIO;
			}
			
			count += inval;
			msg->len += inval;
		}

		if (!(flags & I2C_M_NO_RD_ACK)) {
			inval = acknak(i2c_adap, count);
			if (inval < 0)
				return inval;
		}
	}
	return rdcount;
}

static int bit_doAddress(struct i2c_adapter *i2c_adap, struct i2c_msg *msg)
{
	unsigned short flags = msg->flags;
	unsigned short nak_ok = msg->flags & I2C_M_IGNORE_NAK;
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;

	unsigned char addr;
	int ret, retries;

	retries = nak_ok ? 0 : i2c_adap->retries;

	if (flags & I2C_M_TEN) {
		
		addr = 0xf0 | ((msg->addr >> 7) & 0x03);

		
		ret = try_address(i2c_adap, addr, retries);
		if ((ret != 1) && !nak_ok)  {
			dev_err(&i2c_adap->dev,
				"died at extended address code\n");
			return -EREMOTEIO;
		}
		
		ret = i2c_outb(i2c_adap, msg->addr & 0x7f);
		if ((ret != 1) && !nak_ok) {
			
			dev_err(&i2c_adap->dev, "died at 2nd address code\n");
			return -EREMOTEIO;
		}
		if (flags & I2C_M_RD) {
			bit_dbg(3, &i2c_adap->dev, "emitting repeated "
				"start condition\n");
			i2c_repstart(adap);
			
			addr |= 0x01;
			ret = try_address(i2c_adap, addr, retries);
			if ((ret != 1) && !nak_ok) {
				dev_err(&i2c_adap->dev,
					"died at repeated address code\n");
				return -EREMOTEIO;
			}
		}
	} else {		
		addr = msg->addr << 1;
		if (flags & I2C_M_RD)
			addr |= 1;
		if (flags & I2C_M_REV_DIR_ADDR)
			addr ^= 1;
		ret = try_address(i2c_adap, addr, retries);
		if ((ret != 1) && !nak_ok)
			return -ENXIO;
	}

	return 0;
}

static int bit_xfer(struct i2c_adapter *i2c_adap,
		    struct i2c_msg msgs[], int num)
{
	struct i2c_msg *pmsg;
	struct i2c_algo_bit_data *adap = i2c_adap->algo_data;
	int i, ret;
	unsigned short nak_ok;

    setsda(adap, 0);
    setscl(adap, 0);

	for (i = 0; i < num; i++) {
		pmsg = &msgs[i];
		nak_ok = pmsg->flags & I2C_M_IGNORE_NAK;
		if (!(pmsg->flags & I2C_M_NOSTART)) {
			if (i) {
				bit_dbg(3, &i2c_adap->dev, "emitting "
					"repeated start condition\n");
				i2c_repstart(adap);
			}
			ret = bit_doAddress(i2c_adap, pmsg);
			if ((ret != 0) && !nak_ok) {
				bit_dbg(1, &i2c_adap->dev, "NAK from "
					"device addr 0x%02x msg #%d\n",
					msgs[i].addr, i);
				goto bailout;
			}
		}
		if (pmsg->flags & I2C_M_RD) {
			
			ret = readbytes(i2c_adap, pmsg);

			if (ret < pmsg->len) {
				if (ret >= 0)
					ret = -EREMOTEIO;
				goto bailout;
			}
		} else {
			
			ret = sendbytes(i2c_adap, pmsg);

			if (ret < pmsg->len) {
				if (ret >= 0)
					ret = -EREMOTEIO;
				goto bailout;
			}
		}
	}
	ret = i;

bailout:
	bit_dbg(3, &i2c_adap->dev, "emitting stop condition\n");


    setsda(adap, 0);
    sclhi(adap);
    setsda(adap, 1);

	return ret;
}

static u32 bit_func(struct i2c_adapter *adap)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL |
	       I2C_FUNC_SMBUS_READ_BLOCK_DATA |
	       I2C_FUNC_SMBUS_BLOCK_PROC_CALL |
	       I2C_FUNC_10BIT_ADDR | I2C_FUNC_PROTOCOL_MANGLING;
}

static const struct i2c_algorithm i2c_bit_algo = {
	.master_xfer	= bit_xfer,
	.functionality	= bit_func,
};

static int i2c_bit_prepare_bus(struct i2c_adapter *adap)
{
	struct i2c_algo_bit_data *bit_adap = adap->algo_data;

	if (bit_test) {
		int ret = test_bus(bit_adap, adap->name);
		if (ret < 0)
			return -ENODEV;
	}

	
	adap->algo = &i2c_bit_algo;
	adap->retries = 3;

	return 0;
}

int i2c_cam_bit_add_bus(struct i2c_adapter *adap)
{
	int err;

	err = i2c_bit_prepare_bus(adap);
	if (err)
		return err;

	return i2c_add_adapter(adap);
}
EXPORT_SYMBOL(i2c_cam_bit_add_bus);

int i2c_cam_bit_add_numbered_bus(struct i2c_adapter *adap)
{
	int err;

	err = i2c_bit_prepare_bus(adap);
	if (err)
		return err;

	return i2c_add_numbered_adapter(adap);
}
EXPORT_SYMBOL(i2c_cam_bit_add_numbered_bus);

MODULE_DESCRIPTION("I2C-cam-Bus bit-banging algorithm");

