/*
 * KXTF9 accelerometer driver
 *
 * Copyright (c) 2010 Yamaha Corporation
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2013 KYOCERA Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301, USA.
 */

#include <linux/err.h>
#include <linux/errno.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/i2c.h>
#include <linux/input.h>
#include <linux/uaccess.h>
#include <linux/workqueue.h>
#include <linux/irq.h>
#include <linux/gpio.h>
#include <linux/interrupt.h>
#include <linux/time.h>
#include <linux/slab.h>

#define KXTF9_VERSION "1.2.0"
#define KXTF9_NAME    "kxtf9"

#define DEBUG 0
#define DEBUG_DELAY 0
#define DEBUG_THRESHOLD 0
#define TRACE_FUNC() pr_debug(KXTF9_NAME ": <trace> %s()\n", __FUNCTION__)

#define KXTF9_DBG
/*
 * Default parameters
 */
#define KXTF9_DEFAULT_DELAY               100
#define KXTF9_MAX_DELAY                   2000

/*
 * Registers
 */
#define KXTF9_XOUT_HPF_L                  0x00
#define KXTF9_XOUT_HPF_H                  0x01
#define KXTF9_YOUT_HPF_L                  0x02
#define KXTF9_YOUT_HPF_H                  0x03
#define KXTF9_ZOUT_HPF_L                  0x04
#define KXTF9_ZOUT_HPF_H                  0x05
#define KXTF9_XOUT_L                      0x06
#define KXTF9_XOUT_H                      0x07
#define KXTF9_YOUT_L                      0x08
#define KXTF9_YOUT_H                      0x09
#define KXTF9_ZOUT_L                      0x0a
#define KXTF9_ZOUT_H                      0x0b
#define KXTF9_ACC_REG                     0x06
#define KXTF9_DCST_RESP                   0x0c

#define KXTF9_WHO_AM_I_REG                0x0f
#define KXTF9_WHO_AM_I                    0x01

#define KXTF9_TILT_POS_CUR                0x10
#define KXTF9_TILT_POS_PRE                0x11
#define KXTF9_INT_SRC_REG1                0x15
#define KXTF9_INT_SRC_REG2                0x16
#define KXTF9_STATUS_REG                  0x18
#define KXTF9_INT_REL                     0x1a
#define KXTF9_CTRL_REG1                   0x1b
#define KXTF9_CTRL_REG2                   0x1c
#define KXTF9_CTRL_REG3                   0x1d
#define KXTF9_INT_CTRL_REG1               0x1e
#define KXTF9_INT_CTRL_REG2               0x1f
#define KXTF9_INT_CTRL_REG3               0x20
#define KXTF9_DATA_CTRL_REG               0x21

#define KXTF9_ODR_REG                     0x21
#define KXTF9_ODR_MASK                    0x07
#define KXTF9_ODR_SHIFT                   0
#define KXTF9_ODR_12HZ                    0
#define KXTF9_ODR_25HZ                    1
#define KXTF9_ODR_50HZ                    2
#define KXTF9_ODR_100HZ                   3
#define KXTF9_ODR_200HZ                   4
#define KXTF9_ODR_400HZ                   5
#define KXTF9_ODR_800HZ                   6


#define KXTF9_POWER_CONTROL_REG           0x1b
#define KXTF9_POWER_CONTROL_MASK          0x80
#define KXTF9_POWER_CONTROL_SHIFT         7

#define KXTF9_RES_REG                     0x1b
#define KXTF9_RES_MASK                    0x40
#define KXTF9_RES_SHIFT                   6
#define KXTF9_RES_8                       0
#define KXTF9_RES_12                      1

#define KXTF9_GSEL_REG                    0x1b
#define KXTF9_GSEL_MASK                   0x18
#define KXTF9_GSEL_SHIFT                  3
#define KXTF9_GSEL_2G                     0
#define KXTF9_GSEL_4G                     1
#define KXTF9_GSEL_8G                     2

#define KXTF9_ENABLE_FUNC_REG             0x1b
#define KXTF9_ENABLE_FUNC_MASK            0x07
#define KXTF9_ENABLE_FUNC_SHIFT           0
#define KXTF9_ENABLE_FUNC_TDTE            4
#define KXTF9_ENABLE_FUNC_WUFE            2
#define KXTF9_ENABLE_FUNC_TPE             1

#define KXTF9_SOFT_RESET_REG              0x1d
#define KXTF9_SOFT_RESET_MASK             0x80
#define KXTF9_SOFT_RESET_SHIFT            7

#define KXTF9_TILT_TIMER                  0x28
#define KXTF9_WUF_TIMER                   0x29
#define KXTF9_TDT_TIMER                   0x2b
#define KXTF9_TDT_H_THRESH                0x2c
#define KXTF9_TDT_L_THRESH                0x2d
#define KXTF9_TDT_TAP_TIMER               0x2e
#define KXTF9_TDT_TOTAL_TIMER             0x2f
#define KXTF9_TDT_LATENCY_TIMER           0x30
#define KXTF9_TDT_WINDOW_TIMER            0x31
#define KXTF9_SELF_TEST                   0x3a
#define KXTF9_WUF_THRESH                  0x5a
#define KXTF9_TILT_ANGLE                  0x5c
#define KXTF9_HYST_SET                    0x5f

/*
 * Acceleration measurement
 */
#define KXTF9_RESOLUTION                  64

/* ABS axes parameter range [um/s^2] (for input event) */
#define GRAVITY_EARTH                     9806550
#define ABSMIN_2G                         (-GRAVITY_EARTH * 2)
#define ABSMAX_2G                         (GRAVITY_EARTH * 2)

struct acceleration {
    int x;
    int y;
    int z;
};

/*
 * Output data rate
 */
struct kxtf9_odr {
        unsigned long delay;            /* min delay (msec) in the range of ODR */
        u8 odr;                         /* bandwidth register value */
};

static const struct kxtf9_odr kxtf9_odr_table[] = {
    {1,  KXTF9_ODR_800HZ},
    {2,  KXTF9_ODR_400HZ},
    {5, KXTF9_ODR_200HZ},
    {10, KXTF9_ODR_100HZ},
    {20, KXTF9_ODR_50HZ},
    {40, KXTF9_ODR_25HZ},
};

/*
 * Transformation matrix for chip mounting position
 */
static const int kxtf9_position_map[][3][3] = {
    {{ 1,  0,  0}, { 0,  1,  0}, { 0,  0,  1}}, /* top/upper-left */
    {{ 0,  1,  0}, {-1,  0,  0}, { 0,  0,  1}}, /* top/upper-right */
    {{-1,  0,  0}, { 0, -1,  0}, { 0,  0,  1}}, /* top/lower-right */
    {{ 0, -1,  0}, { 1,  0,  0}, { 0,  0,  1}}, /* top/lower-left */
    {{-1,  0,  0}, { 0,  1,  0}, { 0,  0, -1}}, /* bottom/upper-left */
    {{ 0, -1,  0}, {-1,  0,  0}, { 0,  0, -1}}, /* bottom/upper-right */
    {{ 1,  0,  0}, { 0, -1,  0}, { 0,  0, -1}}, /* bottom/lower-right */
    {{ 0,  1,  0}, { 1,  0,  0}, { 0,  0, -1}}, /* bottom/lower-left */
};

/*
 * driver private data
 */
struct kxtf9_data {
    atomic_t enable;                /* attribute value */
    atomic_t delay;                 /* attribute value */
    atomic_t position;              /* attribute value */
    atomic_t threshold;             /* attribute value */
    struct acceleration last;       /* last measured data */
    struct mutex enable_mutex;
    struct mutex data_mutex;
    struct i2c_client *client;
    struct input_dev *input;
    struct delayed_work work;
#if DEBUG
    int suspend;
#endif
};

#define delay_to_jiffies(d) ((d)?msecs_to_jiffies(d):1)
#define actual_delay(d)     (jiffies_to_msecs(delay_to_jiffies(d)))

/* register access functions */
#define kxtf9_read_bits(p,r) \
    ((i2c_smbus_read_byte_data((p)->client, r##_REG) & r##_MASK) >> r##_SHIFT)
#define kxtf9_update_bits(p,r,v) \
    i2c_smbus_write_byte_data((p)->client, r##_REG, \
                              ((i2c_smbus_read_byte_data((p)->client,r##_REG) & ~r##_MASK) | ((v) << r##_SHIFT)))

/*
 * Device dependant operations
 */

static int kxtf9_power_up(struct kxtf9_data *kxtf9)
{
    kxtf9_update_bits(kxtf9, KXTF9_POWER_CONTROL, 1);

    return 0;
}

static int kxtf9_power_down(struct kxtf9_data *kxtf9)
{
    kxtf9_update_bits(kxtf9, KXTF9_POWER_CONTROL, 0);

    return 0;
}

static int kxtf9_hw_init(struct kxtf9_data *kxtf9)
{
    kxtf9_power_down(kxtf9);

    /* software reset */
    kxtf9_update_bits(kxtf9, KXTF9_SOFT_RESET, 1);
    while (kxtf9_read_bits(kxtf9, KXTF9_SOFT_RESET)) {
        msleep(1);
    }

    /* data output range 8bit */
    kxtf9_update_bits(kxtf9, KXTF9_RES, KXTF9_RES_8);

    /* g range  +/-2g */
    kxtf9_update_bits(kxtf9, KXTF9_GSEL, KXTF9_GSEL_2G);

    /* enable function */
    kxtf9_update_bits(kxtf9, KXTF9_ENABLE_FUNC, KXTF9_ENABLE_FUNC_TDTE | KXTF9_ENABLE_FUNC_WUFE | KXTF9_ENABLE_FUNC_TPE);

    return 0;
}

static int kxtf9_get_enable(struct device *dev)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);

    return atomic_read(&kxtf9->enable);
}

static void kxtf9_set_enable(struct device *dev, int enable)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);
    int delay = atomic_read(&kxtf9->delay);

    mutex_lock(&kxtf9->enable_mutex);

    if (enable) {                   /* enable if state will be changed */
        if (!atomic_cmpxchg(&kxtf9->enable, 0, 1)) {
#ifdef KXTF9_DBG
            printk("kxtf9_set_enable set enable\n");
#endif
            kxtf9_power_up(kxtf9);
            schedule_delayed_work(&kxtf9->work, delay_to_jiffies(delay) + 1);
        }
    } else {                        /* disable if state will be changed */
        if (atomic_cmpxchg(&kxtf9->enable, 1, 0)) {
#ifdef KXTF9_DBG
            printk("kxtf9_set_enable set disable\n");
#endif
            cancel_delayed_work_sync(&kxtf9->work);
            kxtf9_power_down(kxtf9);
        }
    }
    atomic_set(&kxtf9->enable, enable);

    mutex_unlock(&kxtf9->enable_mutex);
}

static int kxtf9_get_delay(struct device *dev)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);

    return atomic_read(&kxtf9->delay);
}

static void kxtf9_set_delay(struct device *dev, int delay)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);
    u8 odr;
    int i;

    /* determine optimum ODR */
    for (i = 1; (i < ARRAY_SIZE(kxtf9_odr_table)) &&
             (actual_delay(delay) >= kxtf9_odr_table[i].delay); i++)
        ;
    odr = kxtf9_odr_table[i-1].odr;
    atomic_set(&kxtf9->delay, delay);

    mutex_lock(&kxtf9->enable_mutex);

    if (kxtf9_get_enable(dev)) {
        cancel_delayed_work_sync(&kxtf9->work);
        kxtf9_power_down(kxtf9);
        kxtf9_update_bits(kxtf9, KXTF9_ODR, odr);
        kxtf9_power_up(kxtf9);
        schedule_delayed_work(&kxtf9->work, delay_to_jiffies(delay) + 1);
    } else {
        kxtf9_update_bits(kxtf9, KXTF9_ODR, odr);
    }

    mutex_unlock(&kxtf9->enable_mutex);
#ifdef KXTF9_DBG
    printk("kxtf9_set_delay Set ODR is 0x%02x\n",odr);
#endif
}

static int kxtf9_get_position(struct device *dev)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);

    return atomic_read(&kxtf9->position);
}

static void kxtf9_set_position(struct device *dev, int position)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);

    atomic_set(&kxtf9->position, position);
}

static int kxtf9_get_threshold(struct device *dev)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);

    return atomic_read(&kxtf9->threshold);
}

static void kxtf9_set_threshold(struct device *dev, int threshold)
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);

    atomic_set(&kxtf9->threshold, threshold);
}

static int kxtf9_data_filter(struct device *dev, struct acceleration *accel, int data[])
{
    struct i2c_client *client = to_i2c_client(dev);
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);
    int threshold = atomic_read(&kxtf9->threshold);
#if DEBUG_THRESHOLD
    int update;
#endif
#if DEBUG_THRESHOLD
    update = 0;
#endif
    mutex_lock(&kxtf9->data_mutex);
    if ((abs(kxtf9->last.x - data[0]) > threshold) ||
        (abs(kxtf9->last.y - data[1]) > threshold) ||
        (abs(kxtf9->last.z - data[2]) > threshold)) {
        accel->x = data[0];
        accel->y = data[1];
        accel->z = data[2];
#if DEBUG_THRESHOLD
        update = 1;
#endif
    } else {
        *accel = kxtf9->last;
    }

#if DEBUG_THRESHOLD
    if (update == 1) {
        dev_info(&client->dev, "threshold=%d x(%d) y(%d) z(%d) accel(%d,%d,%d) ****\n", threshold,
                 kxtf9->last.x - data[0], kxtf9->last.y - data[1], kxtf9->last.z - data[2], accel->x, accel->y, accel->z);
    } else {
        dev_info(&client->dev, "threshold=%d x(%d) y(%d) z(%d) accel(%d,%d,%d)\n", threshold,
                 kxtf9->last.x - data[0], kxtf9->last.y - data[1], kxtf9->last.z - data[2], accel->x, accel->y, accel->z);
    }
#endif
    mutex_unlock(&kxtf9->data_mutex);

    return 0;
}

static int kxtf9_measure(struct kxtf9_data *kxtf9, struct acceleration *accel)
{
    struct i2c_client *client = kxtf9->client;
    u8 buf[6];
    int raw[3], data[3];
    int pos = atomic_read(&kxtf9->position);
    long long g;
    int i, j;
#if DEBUG_DELAY
    struct timespec t;
#endif

#if DEBUG_DELAY
    getnstimeofday(&t);
#endif

    /* read acceleration data */
    if (i2c_smbus_read_i2c_block_data(client, KXTF9_ACC_REG, 6, buf) != 6) {
        dev_err(&client->dev,
                "I2C block read error: addr=0x%02x, len=%d\n",
                KXTF9_ACC_REG, 6);
            raw[0] = 0;
            raw[1] = 0;
            raw[2] = 0;
    } else {
        raw[0] = (s8)(buf[1]);
        raw[1] = (s8)(buf[3]);
        raw[2] = (s8)(buf[5]);
    }

    /* for X, Y, Z axis */
    for (i = 0; i < 3; i++) {
        /* coordinate transformation */
        data[i] = 0;
        for (j = 0; j < 3; j++) {
            data[i] += raw[j] * kxtf9_position_map[pos][i][j];
        }
        /* normalization */
        g = (long long)data[i] * GRAVITY_EARTH / KXTF9_RESOLUTION;
        data[i] = g;
    }

    dev_dbg(&client->dev, "raw(%5d,%5d,%5d) => norm(%8d,%8d,%8d)\n",
            raw[0], raw[1], raw[2], data[0], data[1], data[2]);

#if DEBUG_DELAY
    dev_info(&client->dev, "%ld.%lds:raw(%5d,%5d,%5d) => norm(%8d,%8d,%8d)\n", t.tv_sec, t.tv_nsec,
             raw[0], raw[1], raw[2], data[0], data[1], data[2]);
#endif

    kxtf9_data_filter(&client->dev, accel, data);

    return 0;
}

#ifdef KXTF9_DBG
static char dbg_show_cnt = 0;
#endif

static void kxtf9_work_func(struct work_struct *work)
{
    struct kxtf9_data *kxtf9 = container_of((struct delayed_work *)work, struct kxtf9_data, work);
    struct acceleration accel;
    unsigned long delay = delay_to_jiffies(atomic_read(&kxtf9->delay));

#ifdef KXTF9_DBG
    if (!dbg_show_cnt) printk("kxtf9_work_func start\n");
#endif
    kxtf9_measure(kxtf9, &accel);
#ifdef KXTF9_DBG
    if (!dbg_show_cnt) {
        printk("kxtf9_work_func accel data(dec) x:%d, y:%d, z:%d\n",accel.x,accel.y,accel.z);
        printk("kxtf9_work_func accel data(hex) x:0x%04x, y:0x%04x, z:0x%04x\n",accel.x,accel.y,accel.z);
    }
#endif

    input_report_abs(kxtf9->input, ABS_X, accel.x);
    input_report_abs(kxtf9->input, ABS_Y, accel.y);
    input_report_abs(kxtf9->input, ABS_Z, accel.z);
    input_sync(kxtf9->input);

    mutex_lock(&kxtf9->data_mutex);
    kxtf9->last = accel;
    mutex_unlock(&kxtf9->data_mutex);

    schedule_delayed_work(&kxtf9->work, delay);
#ifdef KXTF9_DBG
    if (!dbg_show_cnt) printk("kxtf9_work_func end\n");
    dbg_show_cnt++;
    if (dbg_show_cnt >= 100) dbg_show_cnt = 0;
#endif
}

/*
 * Input device interface
 */
static int kxtf9_input_init(struct kxtf9_data *kxtf9)
{
    struct input_dev *dev;
    int err;

    dev = input_allocate_device();
    if (!dev) {
#ifdef KXTF9_DBG
        printk("kxtf9_input_init alloc device err\n");
#endif
        return -ENOMEM;
    }
    dev->name = "accelerometer";
    dev->id.bustype = BUS_I2C;

    input_set_capability(dev, EV_ABS, ABS_MISC);
    input_set_abs_params(dev, ABS_X, ABSMIN_2G, ABSMAX_2G, 0, 0);
    input_set_abs_params(dev, ABS_Y, ABSMIN_2G, ABSMAX_2G, 0, 0);
    input_set_abs_params(dev, ABS_Z, ABSMIN_2G, ABSMAX_2G, 0, 0);
    input_set_drvdata(dev, kxtf9);

    err = input_register_device(dev);
    if (err < 0) {
#ifdef KXTF9_DBG
        printk("kxtf9_input_init input register device err\n");
#endif
        input_free_device(dev);
        return err;
    }
    kxtf9->input = dev;

    return 0;
}

static void kxtf9_input_fini(struct kxtf9_data *kxtf9)
{
    struct input_dev *dev = kxtf9->input;

    input_unregister_device(dev);
    input_free_device(dev);
}

/*
 * sysfs device attributes
 */
static ssize_t kxtf9_enable_show(struct device *dev,
                                 struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%d\n", kxtf9_get_enable(dev));
}

static ssize_t kxtf9_enable_store(struct device *dev,
                                  struct device_attribute *attr,
                                  const char *buf, size_t count)
{
    unsigned long enable = simple_strtoul(buf, NULL, 10);

#ifdef KXTF9_DBG
    printk("kxtf9_enable_store : %d\n",(int)enable);
#endif

    if ((enable == 0) || (enable == 1)) {
        kxtf9_set_enable(dev, enable);
    }

    return count;
}

static ssize_t kxtf9_delay_show(struct device *dev,
                                struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%d\n", kxtf9_get_delay(dev));
}

static ssize_t kxtf9_delay_store(struct device *dev,
                                 struct device_attribute *attr,
                                 const char *buf, size_t count)
{
    unsigned long delay = simple_strtoul(buf, NULL, 10);

#ifdef KXTF9_DBG
    printk("kxtf9_delay_store : %d\n",(int)delay);
#endif

    if (delay > KXTF9_MAX_DELAY) {
        delay = KXTF9_MAX_DELAY;
    }

    kxtf9_set_delay(dev, delay);

   return count;
}

static ssize_t kxtf9_position_show(struct device *dev,
                                   struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%d\n", kxtf9_get_position(dev));
}

static ssize_t kxtf9_position_store(struct device *dev,
                                    struct device_attribute *attr,
                                    const char *buf, size_t count)
{
    unsigned long position;

    position = simple_strtoul(buf, NULL,10);
#ifdef KXTF9_DBG
    printk("kxtf9_position_store : %d\n",(int)position);
#endif
    if ((position >= 0) && (position <= 7)) {
        kxtf9_set_position(dev, position);
    }

    return count;
}

static ssize_t kxtf9_threshold_show(struct device *dev,
                                    struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%d\n", kxtf9_get_threshold(dev));
}

static ssize_t kxtf9_threshold_store(struct device *dev,
                                     struct device_attribute *attr,
                                     const char *buf, size_t count)
{
    unsigned long threshold;

    threshold = simple_strtoul(buf, NULL,10);
#ifdef KXTF9_DBG
    printk("kxtf9_threshold_store : %d\n",(int)threshold);
#endif
    if (threshold >= 0 && threshold <= ABSMAX_2G) {
        kxtf9_set_threshold(dev, threshold);
    }

    return count;
}

static ssize_t kxtf9_wake_store(struct device *dev,
                                struct device_attribute *attr,
                                const char *buf, size_t count)
{
    struct input_dev *input = to_input_dev(dev);
    static atomic_t serial = ATOMIC_INIT(0);

#ifdef KXTF9_DBG
    printk("kxtf9_wake_store\n");
#endif
    input_report_abs(input, ABS_MISC, atomic_inc_return(&serial));

    return count;
}

static ssize_t kxtf9_data_show(struct device *dev,
                               struct device_attribute *attr, char *buf)
{
    struct input_dev *input = to_input_dev(dev);
    struct kxtf9_data *kxtf9 = input_get_drvdata(input);
    struct acceleration accel;

    mutex_lock(&kxtf9->data_mutex);
    accel = kxtf9->last;
    mutex_unlock(&kxtf9->data_mutex);

    return sprintf(buf, "%d %d %d\n", accel.x, accel.y, accel.z);
}

#if DEBUG
static ssize_t kxtf9_debug_reg_show(struct device *dev,
                                     struct device_attribute *attr, char *buf)
{
    struct input_dev *input = to_input_dev(dev);
    struct kxtf9_data *kxtf9 = input_get_drvdata(input);
    struct i2c_client *client = kxtf9->client;
    ssize_t count = 0;
    u8 reg;

    reg = i2c_smbus_read_byte_data(client, KXTF9_XOUT_HPF_L);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_XOUT_HPF_L, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_XOUT_HPF_H);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_XOUT_HPF_H, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_YOUT_HPF_L);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_YOUT_HPF_L, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_YOUT_HPF_H);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_YOUT_HPF_H, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_ZOUT_HPF_L);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_ZOUT_HPF_L, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_ZOUT_HPF_H);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_ZOUT_HPF_H, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_XOUT_L);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_XOUT_L, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_XOUT_H);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_XOUT_H, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_YOUT_L);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_YOUT_L, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_YOUT_H);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_YOUT_H, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_ZOUT_L);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_ZOUT_L, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_ZOUT_H);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_ZOUT_H, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_DCST_RESP);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_DCST_RESP, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_WHO_AM_I);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_WHO_AM_I, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_TILT_POS_CUR);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_TILT_POS_CUR, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_TILT_POS_PRE);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_TILT_POS_PRE, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_INT_SRC_REG1);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_INT_SRC_REG1, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_INT_SRC_REG2);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_INT_SRC_REG2, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_STATUS_REG);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_STATUS_REG, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_INT_REL);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_INT_REL, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_CTRL_REG1);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_CTRL_REG1, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_CTRL_REG2);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_CTRL_REG2, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_CTRL_REG3);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_CTRL_REG3, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_INT_CTRL_REG1);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_INT_CTRL_REG1, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_INT_CTRL_REG2);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_INT_CTRL_REG2, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_INT_CTRL_REG3);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_INT_CTRL_REG3, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_DATA_CTRL_REG);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_DATA_CTRL_REG, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_TILT_TIMER);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_TILT_TIMER, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_WUF_TIMER);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_WUF_TIMER, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_TDT_TIMER);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_TDT_TIMER, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_TDT_H_THRESH);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_TDT_H_THRESH, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_TDT_L_THRESH);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_TDT_L_THRESH, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_TDT_TAP_TIMER);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_TDT_TAP_TIMER, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_TDT_TOTAL_TIMER);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_TDT_TOTAL_TIMER, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_TDT_LATENCY_TIMER);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_TDT_LATENCY_TIMER, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_TDT_WINDOW_TIMER);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_TDT_WINDOW_TIMER, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_SELF_TEST);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_SELF_TEST, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_WUF_THRESH);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_WUF_THRESH, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_TILT_ANGLE);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_TILT_ANGLE, reg);
    reg = i2c_smbus_read_byte_data(client, KXTF9_HYST_SET);
    count += sprintf(&buf[count], "%02x: %d\n", KXTF9_HYST_SET, reg);

    return count;
}

static int kxtf9_suspend(struct i2c_client *client, pm_message_t mesg);
static int kxtf9_resume(struct i2c_client *client);

static ssize_t kxtf9_debug_suspend_show(struct device *dev,
                                        struct device_attribute *attr, char *buf)
{
    struct input_dev *input = to_input_dev(dev);
    struct kxtf9_data *kxtf9 = input_get_drvdata(input);

    return sprintf(buf, "%d\n", kxtf9->suspend);
}

static ssize_t kxtf9_debug_suspend_store(struct device *dev,
                                         struct device_attribute *attr,
                                         const char *buf, size_t count)
{
    struct input_dev *input = to_input_dev(dev);
    struct kxtf9_data *kxtf9 = input_get_drvdata(input);
    struct i2c_client *client = kxtf9->client;
    unsigned long suspend = simple_strtoul(buf, NULL, 10);

    if (suspend) {
        pm_message_t msg;
        kxtf9_suspend(client, msg);
    } else {
        kxtf9_resume(client);
    }

    return count;
}
#endif /* DEBUG */

static DEVICE_ATTR(enable, S_IRUGO|S_IWUSR|S_IWGRP,
                   kxtf9_enable_show, kxtf9_enable_store);
static DEVICE_ATTR(delay, S_IRUGO|S_IWUSR|S_IWGRP,
                   kxtf9_delay_show, kxtf9_delay_store);
static DEVICE_ATTR(position, S_IRUGO|S_IWUSR,
                   kxtf9_position_show, kxtf9_position_store);
static DEVICE_ATTR(threshold, S_IRUGO|S_IWUSR,
                   kxtf9_threshold_show, kxtf9_threshold_store);
static DEVICE_ATTR(wake, S_IWUSR|S_IWGRP, NULL, kxtf9_wake_store);
static DEVICE_ATTR(data, S_IRUGO,
                   kxtf9_data_show, NULL);

#if DEBUG
static DEVICE_ATTR(debug_reg, S_IRUGO,
                   kxtf9_debug_reg_show, NULL);
static DEVICE_ATTR(debug_suspend, S_IRUGO|S_IWUSR,
                   kxtf9_debug_suspend_show, kxtf9_debug_suspend_store);
#endif /* DEBUG */

static struct attribute *kxtf9_attributes[] = {
    &dev_attr_enable.attr,
    &dev_attr_delay.attr,
    &dev_attr_position.attr,
    &dev_attr_threshold.attr,
    &dev_attr_wake.attr,
    &dev_attr_data.attr,
#if DEBUG
    &dev_attr_debug_reg.attr,
    &dev_attr_debug_suspend.attr,
#endif /* DEBUG */
    NULL
};

static struct attribute_group kxtf9_attribute_group = {
    .attrs = kxtf9_attributes
};

/*
 * I2C client
 */
static int kxtf9_detect(struct i2c_client *client, struct i2c_board_info *info)
{
    int id;

#ifdef KXTF9_DBG
    printk("kxtf9_detect start\n");
#endif
    id = i2c_smbus_read_byte_data(client, KXTF9_WHO_AM_I_REG);
    if (id != KXTF9_WHO_AM_I) {
#ifdef KXTF9_DBG
        printk("kxtf9_detect i2c read err\n");
#endif
        return -ENODEV;
    }
#ifdef KXTF9_DBG
    printk("kxtf9_detect end\n");
#endif
    return 0;
}


static int kxtf9_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
    struct kxtf9_data *kxtf9;
    int err;

    TRACE_FUNC();

#ifdef KXTF9_DBG
    printk("kxtf9_probe start\n");
#endif
    /* setup private data */
    kxtf9 = kzalloc(sizeof(struct kxtf9_data), GFP_KERNEL);
    if (!kxtf9) {
#ifdef KXTF9_DBG
        printk("kxtf9_probe kzalloc err\n");
#endif
        err = -ENOMEM;
        goto error_0;
    }

    mutex_init(&kxtf9->enable_mutex);
    mutex_init(&kxtf9->data_mutex);

    /* setup i2c client */
    if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
#ifdef KXTF9_DBG
        printk("kxtf9_probe i2c functionality err\n");
#endif
        err = -ENODEV;
        goto error_1;
    }

    i2c_set_clientdata(client, kxtf9);
    kxtf9->client = client;

    /* detect and init hardware */
    if ((err = kxtf9_detect(client, NULL))) {
#ifdef KXTF9_DBG
        printk("kxtf9_probe detect err\n");
#endif
        goto error_1;
    }
    dev_info(&client->dev, "%s found\n", id->name);

    kxtf9_hw_init(kxtf9);
    kxtf9_set_delay(&client->dev, KXTF9_DEFAULT_DELAY);
    kxtf9_set_position(&client->dev, CONFIG_INPUT_KXTF9_POSITION);

    /* setup driver interfaces */
    INIT_DELAYED_WORK(&kxtf9->work, kxtf9_work_func);

    err = kxtf9_input_init(kxtf9);
    if (err < 0) {
#ifdef KXTF9_DBG
        printk("kxtf9_probe init err\n");
#endif
        goto error_1;
    }

    err = sysfs_create_group(&kxtf9->input->dev.kobj, &kxtf9_attribute_group);
    if (err < 0) {
#ifdef KXTF9_DBG
        printk("kxtf9_probe sysfs create err\n");
#endif
        goto error_2;
    }

#ifdef KXTF9_DBG
    printk("kxtf9_probe end\n");
#endif
    return 0;

error_2:
    kxtf9_input_fini(kxtf9);
error_1:
    kfree(kxtf9);
error_0:
#ifdef KXTF9_DBG
    printk("kxtf9_probe err end\n");
#endif
    return err;
}

static int kxtf9_remove(struct i2c_client *client)
{
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);

#ifdef KXTF9_DBG
    printk("kxtf9_remove start\n");
#endif
    kxtf9_set_enable(&client->dev, 0);

    sysfs_remove_group(&kxtf9->input->dev.kobj, &kxtf9_attribute_group);
    kxtf9_input_fini(kxtf9);

    kfree(kxtf9);

#ifdef KXTF9_DBG
    printk("kxtf9_remove end\n");
#endif
    return 0;
}

static int kxtf9_suspend(struct i2c_client *client, pm_message_t mesg)
{
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);

#ifdef KXTF9_DBG
    printk("kxtf9_suspend start\n");
#endif
    mutex_lock(&kxtf9->enable_mutex);

    if (kxtf9_get_enable(&client->dev)) {
        cancel_delayed_work_sync(&kxtf9->work);
        kxtf9_power_down(kxtf9);
    }

#if DEBUG
    kxtf9->suspend = 1;
#endif

    mutex_unlock(&kxtf9->enable_mutex);

#ifdef KXTF9_DBG
    printk("kxtf9_suspend end\n");
#endif
    return 0;
}

static int kxtf9_resume(struct i2c_client *client)
{
    struct kxtf9_data *kxtf9 = i2c_get_clientdata(client);
    int delay = atomic_read(&kxtf9->delay);

#ifdef KXTF9_DBG
    printk("kxtf9_resume start\n");
#endif
    kxtf9_hw_init(kxtf9);
    kxtf9_set_delay(&client->dev, delay);

    mutex_lock(&kxtf9->enable_mutex);

    if (kxtf9_get_enable(&client->dev)) {
        kxtf9_power_up(kxtf9);
        schedule_delayed_work(&kxtf9->work, delay_to_jiffies(delay) + 1);
    }

#if DEBUG
    kxtf9->suspend = 0;
#endif

    mutex_unlock(&kxtf9->enable_mutex);

#ifdef KXTF9_DBG
    printk("kxtf9_resume end\n");
#endif
    return 0;
}

static const struct i2c_device_id kxtf9_id[] = {
    {KXTF9_NAME, 0},
    {},
};

MODULE_DEVICE_TABLE(i2c, kxtf9_id);

static struct i2c_driver kxtf9_driver = {
    .driver = {
        .name = KXTF9_NAME,
        .owner = THIS_MODULE,
    },
    .probe = kxtf9_probe,
    .remove = kxtf9_remove,
    .suspend = kxtf9_suspend,
    .resume = kxtf9_resume,
    .id_table = kxtf9_id,
};

static int __init kxtf9_init(void)
{
#if defined(CONFIG_TARGET_PROTOTYPE_WS0)
    return i2c_add_driver(&kxtf9_driver);
#else
    return -1;
#endif
}

static void __exit kxtf9_exit(void)
{
    i2c_del_driver(&kxtf9_driver);
}

module_init(kxtf9_init);
module_exit(kxtf9_exit);

MODULE_DESCRIPTION("KXTF9 accelerometer driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(KXTF9_VERSION);
